from enum import Enum


REGIONS = ['br', 'cn', 'eun', 'eune', 'euw', 'garena2', 'garena3', 'id', 'jp', 'kr', 'la', 'la1', 'la2', 'lan', 'las',
           'na', 'oc', 'oc1', 'oce', 'pbe', 'ph', 'ph2', 'ru', 'sg', 'tencent', 'th', 'th2', 'tr', 'tw', 'tw2', 'vn', 'vn2']

class Language(Enum):
    ar_ae = 'ar_ae'
    ar_eg = 'ar_eg'
    cs_cz = 'cs_cz'
    de_de = 'de_de'
    el_gr = 'el_gr'
    en_au = 'en_au'
    en_gb = 'en_gb'
    en_ph = 'en_ph'
    en_pl = 'en_pl'
    en_sg = 'en_sg'
    en_us = 'en_us'
    es_ar = 'es_ar'
    es_es = 'es_es'
    es_mx = 'es_mx'
    fr_fr = 'fr_fr'
    hu_hu = 'hu_hu'
    id_id = 'id_id'
    it_it = 'it_it'
    ja_jp = 'ja_jp'
    ko_kr = 'ko_kr'
    ms_my = 'ms_my'
    pl_pl = 'pl_pl'
    pt_br = 'pt_br'
    ro_ro = 'ro_ro'
    ru_ru = 'ru_ru'
    th_th = 'th_th'
    tr_tr = 'tr_tr'
    vi_vn = 'vi_vn'
    vn_vn = 'vn_vn'
    zh_cn = 'zh_cn'
    zh_my = 'zh_my'
    zh_tw = 'zh_tw'
